import pathlib
from os import remove

from utils import read_stories_from_excel

from jira_assistant.excel_definition import ExcelDefinition
from jira_assistant.excel_operation import output_to_excel_file, read_excel_file
from jira_assistant.sprint_schedule import SprintScheduleStore

HERE = pathlib.Path(__file__).resolve().parent
SRC_ASSETS: pathlib.Path = HERE.parent / "src/jira_assistant/assets"


class TestExcelOperation:
    def test_read_excel_file(self):
        excel_definition = ExcelDefinition()
        excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
        sprint_schedule = SprintScheduleStore()
        sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

        columns, stories = read_excel_file(
            HERE / "files/happy_path.xlsx", excel_definition, sprint_schedule
        )
        assert len(columns) == 24
        assert len(stories) == 8

    def test_output_to_excel_file(self):
        stories = read_stories_from_excel(
            HERE / "files/happy_path.xlsx",
            SRC_ASSETS / "excel_definition.json",
            SRC_ASSETS / "sprint_schedule.json",
        )

        output_to_excel_file(
            HERE / "files/happy_path_direct_output.xlsx",
            stories,
            ExcelDefinition().load_file(SRC_ASSETS / "excel_definition.json"),
        )

        assert (HERE / "files/happy_path_direct_output.xlsx").exists()

        remove(HERE / "files/happy_path_direct_output.xlsx")



from decimal import Decimal

from pytest import raises
from utils import mock_story_data

from jira_assistant.story import (
    compare_story_based_on_inline_weights,
    convert_to_bool,
    convert_to_datetime,
    convert_to_decimal,
)


class TestStory:
    def test_compare_story_based_on_inline_weight(self):
        data = mock_story_data()
        s_1 = data[0]
        s_2 = data[1]
        s_3 = data[2]
        s_4 = data[3]
        s_5 = data[4]
        s_6 = data[5]
        s_7 = data[6]
        s_8 = data[7]
        s_9 = data[8]
        assert compare_story_based_on_inline_weights(s_1, s_2) < 0
        assert compare_story_based_on_inline_weights(s_1, s_3) < 0
        assert compare_story_based_on_inline_weights(s_2, s_3) < 0
        assert compare_story_based_on_inline_weights(s_3, s_5) < 0
        assert compare_story_based_on_inline_weights(s_2, s_5) < 0
        assert compare_story_based_on_inline_weights(s_4, s_5) < 0
        assert compare_story_based_on_inline_weights(s_1, s_5) < 0
        assert compare_story_based_on_inline_weights(s_6, s_7) < 0
        assert compare_story_based_on_inline_weights(s_8, s_9) < 0

    def test_compare_story_property(self):
        data = mock_story_data()
        s_8 = data[7]
        s_9 = data[8]
        assert s_8["name"] < s_9["name"]
        assert s_8["productValue"] < s_9["productValue"]

    def test_str(self):
        data = mock_story_data()

        assert "s1, N/A" in str(data[0])

    def test_lt_le_gt_ge_eq(self):
        data = mock_story_data()
        s_1 = data[0]
        s_2 = data[1]

        with raises(TypeError):
            assert s_1 < s_2
        with raises(TypeError):
            assert s_1 <= s_2
        with raises(TypeError):
            assert s_1 > s_2
        with raises(TypeError):
            assert s_1 >= s_2
        with raises(TypeError):
            assert s_1 == s_2

    def test_convert_to_bool_using_correct_type(self):
        assert convert_to_bool(True) is True

    def test_convert_to_decimal(self):
        assert Decimal.compare(
            convert_to_decimal(Decimal(1.2)), Decimal(1.2)
        ) == Decimal("0")
        assert Decimal.compare(convert_to_decimal("1.2"), Decimal("1.2")) == Decimal(
            "0"
        )
        assert Decimal.compare(convert_to_decimal("good"), Decimal(0)) == Decimal("0")

    def test_convert_to_datetime(self):
        assert convert_to_datetime(None) is None



import pathlib
from os import remove, walk
from subprocess import CalledProcessError, run

import pytest

HERE = pathlib.Path(__file__).resolve().parent


class TestConsoleScript:
    # Avoid call JIRA server.
    def test_process_excel_file(self):
        result = run(
            [
                "process-excel-file",
                HERE / "files/happy_path.xlsx",
                "--excel_definition_file",
                HERE / "files/excel_definition_avoid_jira_operations.json",
                "--sprint_schedule_file",
                HERE / "files/sprint_schedule.json",
            ],
            capture_output=True,
            check=True,
        )

        assert "xlsx has been saved" in result.stdout.decode("utf-8")

        remove(HERE / "files/happy_path_sorted.xlsx")

    def test_process_excel_file_apply_output_folder(self):
        result = run(
            [
                "process-excel-file",
                HERE / "files/happy_path.xlsx",
                "--excel_definition_file",
                HERE / "files/excel_definition_avoid_jira_operations.json",
                "--sprint_schedule_file",
                HERE / "files/sprint_schedule.json",
                "--output_folder",
                HERE / "temp",
            ],
            capture_output=True,
            check=True,
        )

        assert "xlsx has been saved" in result.stdout.decode("utf-8")
        assert (HERE / "temp/happy_path_sorted.xlsx").exists()

        remove(HERE / "temp/happy_path_sorted.xlsx")

    def test_process_excel_file_run_twice(self):
        result = run(
            [
                "process-excel-file",
                HERE / "files/happy_path.xlsx",
                "--excel_definition_file",
                HERE / "files/excel_definition_avoid_jira_operations.json",
                "--sprint_schedule_file",
                HERE / "files/sprint_schedule.json",
            ],
            capture_output=True,
            check=True,
        )

        assert "xlsx has been saved" in result.stdout.decode("utf-8")
        assert (HERE / "files/happy_path_sorted.xlsx").exists()

        result = run(
            [
                "process-excel-file",
                HERE / "files/happy_path.xlsx",
                "--excel_definition_file",
                HERE / "files/excel_definition_avoid_jira_operations.json",
                "--sprint_schedule_file",
                HERE / "files/sprint_schedule.json",
            ],
            capture_output=True,
            check=True,
        )

        assert "xlsx has been saved" in result.stdout.decode("utf-8")
        assert (HERE / "files/happy_path_sorted_1.xlsx").exists()

        remove(HERE / "files/happy_path_sorted.xlsx")
        remove(HERE / "files/happy_path_sorted_1.xlsx")

    def test_process_excel_file_using_invalid_definition_file(self):
        with pytest.raises(CalledProcessError) as e:
            run(
                [
                    "process-excel-file",
                    HERE / "files/happy_path.xlsx",
                    "--excel_definition_file",
                    HERE / "files/excel_definition_invalid_structure.txt",
                    "--sprint_schedule_file",
                    HERE / "files/sprint_schedule.json",
                ],
                capture_output=True,
                check=True,
            )

        assert "The structure of excel definition file is wrong" in str(e.value.output)

    def test_process_excel_file_using_wrong_input_file(self):
        with pytest.raises(CalledProcessError) as e:
            run(
                [
                    "process-excel-file",
                    HERE / "files/excel_definition_invalid_index.json",
                    "--excel_definition_file",
                    HERE / "files/excel_definition_invalid_structure.txt",
                ],
                capture_output=True,
                check=True,
            )

        assert "Please provide an Excel file" in str(e.value.output)

    def test_process_excel_file_input_file_not_found(self):
        with pytest.raises(CalledProcessError) as e:
            run(
                [
                    "process-excel-file",
                    HERE / "files/not_found.xlsx",
                    "--excel_definition_file",
                    HERE / "files/excel_definition_invalid_structure.txt",
                ],
                capture_output=True,
                check=True,
            )

        assert "Input file is not exist" in str(e.value.output)

    def test_generate_template_excel_definition(self):
        result = run(
            ["generate-template", "excel-definition"], capture_output=True, check=True
        )

        assert "Generate success" in result.stdout.decode("utf-8")
        assert "excel-definition" in result.stdout.decode("utf-8")

    def test_generate_template_excel(self):
        result = run(["generate-template", "excel"], capture_output=True, check=True)

        assert "Generate success" in result.stdout.decode("utf-8")
        assert "excel" in result.stdout.decode("utf-8")

    def test_generate_template_sprint_schedule(self):
        result = run(
            ["generate-template", "sprint-schedule"], capture_output=True, check=True
        )

        assert "Generate success" in result.stdout.decode("utf-8")
        assert "schedule" in result.stdout.decode("utf-8")

    def test_generate_template_failed(self):
        with pytest.raises(CalledProcessError) as e:
            run(["generate-template", "abc"], capture_output=True, check=True)

        assert "argument template_type: invalid choice" in str(e.value.stderr)

    def test_update_jira_info(self):
        result = run(
            [
                "update-jira-info",
                "--access_token",
                "123456",
                "--url",
                "http://localhost",
            ],
            capture_output=True,
            check=True,
        )

        assert "Add/Update jira url success" in result.stdout.decode("utf-8")
        assert "Add/Update jira access token success" in result.stdout.decode("utf-8")

    def test_update_jira_info_failed(self):
        with pytest.raises(CalledProcessError) as e:
            run(
                ["update-jira-info", "--access_token", " "],
                capture_output=True,
                check=True,
            )

        assert "Please check the access token" in str(e.value.output)

    def teardown_method(self):
        for _, _, files in walk(pathlib.Path.cwd().absolute(), topdown=False):
            for file in files:
                if file.startswith("excel-definition") and "template" in file:
                    remove(file)
                if file.startswith("excel-template"):
                    remove(file)
                if file.startswith("sprint-schedule"):
                    remove(file)



from __future__ import annotations

from pathlib import Path
from sys import version_info
from typing import List, TypedDict, Union

if version_info < (3, 11):
    from typing_extensions import NotRequired, Self
else:
    from typing import NotRequired, Self


def is_absolute_path_valid(path: Union[str, Path]) -> bool:
    if path is None or not Path(path).is_absolute() or not Path(path).exists():
        return False
    return True


class JiraFieldTypeDefinition(TypedDict):
    type: str
    name: NotRequired[str]
    properties: NotRequired[List[Self]]
    isBasic: NotRequired[bool]
    isArray: NotRequired[bool]
    itemType: NotRequired[str]


_jira_field_types: List[JiraFieldTypeDefinition] = [
    {
        "type": "any",
        "isBasic": True,
        "isArray": False,
    },
    {
        "type": "array",
        "isBasic": False,
        "isArray": True,
    },
    {
        "type": "date",
        "isBasic": True,
        "isArray": False,
    },
    {
        "type": "datetime",
        "isBasic": True,
        "isArray": False,
    },
    {
        "type": "string",
        "isBasic": True,
        "isArray": False,
    },
    {
        "type": "number",
        "isBasic": True,
        "isArray": False,
    },
    {
        "type": "user",
        "properties": [
            {"name": "name", "type": "string"},
            {"name": "emailAddress", "type": "string"},
            {"name": "displayName", "type": "string"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "securitylevel",
        "properties": [{"name": "name", "type": "string"}],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "priority",
        "properties": [{"name": "name", "type": "string"}],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "watches",
        "properties": [{"name": "watchCount", "type": "number"}],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "comments-page",
        "properties": [
            {
                "name": "comments",
                "itemType": "comment",
                "type": "array",
                "isArray": True,
            }
        ],
        "isArray": False,
        "isBasic": False,
    },
    {
        "type": "comment",
        "properties": [
            {"name": "author", "type": "author"},
            {"name": "id", "type": "string"},
            {"name": "body", "type": "string"},
            {"name": "created", "type": "datetime"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "author",
        "properties": [
            {"name": "name", "type": "string"},
            {"name": "emailAddress", "type": "string"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "status",
        "properties": [
            {"name": "name", "type": "string"},
            {"name": "statusCategory", "type": "statusCategory"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "statusCategory",
        "properties": [
            {"name": "key", "type": "string"},
            {"name": "name", "type": "string"},
        ],
        "isArray": False,
        "isBasic": False,
    },
    {
        "type": "progress",
        "properties": [
            {"name": "progress", "type": "number"},
            {"name": "total", "type": "number"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "timetracking",
        "properties": [
            {"name": "timeSpent", "type": "string"},
            {"name": "timeSpentSeconds", "type": "number"},
            {"name": "remainingEstimate", "type": "string"},
            {"name": "remainingEstimateSeconds", "type": "number"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "option",
        "properties": [{"name": "value", "type": "string"}],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "resolution",
        "properties": [
            {"name": "description", "type": "string"},
            {"name": "name", "type": "string"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "issuetype",
        "properties": [
            {"name": "description", "type": "string"},
            {"name": "name", "type": "string"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "project",
        "properties": [
            {"name": "name", "type": "string"},
            {"name": "key", "type": "string"},
            {"name": "projectTypeKey", "type": "string"},
            {"name": "projectCategory", "type": "projectCategory"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {
        "type": "projectCategory",
        "properties": [
            {"name": "description", "type": "string"},
            {"name": "name", "type": "string"},
        ],
        "isBasic": False,
        "isArray": False,
    },
    {"type": "issuelinks", "itemType": "issuelink", "isArray": True, "isBasic": False},
    {
        "type": "issuelink",
        "properties": [{"name": "id", "type": "string"}],
        "isBasic": False,
        "isArray": False,
    },
]


def get_jira_field_type(type_name: str | None) -> JiraFieldTypeDefinition | None:
    if type_name is None:
        return None
    for field_type in _jira_field_types:
        if field_type["type"].lower() == type_name.lower():
            return field_type
    return None



import pathlib

from jira_assistant.sprint_schedule import SprintScheduleStore

HERE = pathlib.Path(__file__).resolve().parent
SRC_ASSETS = HERE.parent / "src/jira_assistant/assets"


class TestSprintSchedule:
    def test_load_sprint_schedule(self):
        schedule_filename = SRC_ASSETS / "sprint_schedule.json"
        store = SprintScheduleStore()
        store.load_file(schedule_filename)
        assert store.total_count() > 0

    def test_get_priority(self):
        schedule_filename = SRC_ASSETS / "sprint_schedule.json"
        store = SprintScheduleStore()
        assert store.get_priority("R138") == 0

        store.load_file(schedule_filename)
        assert store.get_priority("R140") == 2



import pathlib
import warnings

try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

from json import dump
from os import environ, remove
from pathlib import Path
from typing import List, Optional, Union

from dotenv import load_dotenv
from urllib3 import disable_warnings

from .excel_definition import ExcelDefinition
from .excel_operation import output_to_excel_file, read_excel_file
from .jira_client import JiraClient
from .sprint_schedule import SprintScheduleStore
from .story import (
    Story,
    sort_stories_by_inline_weights,
    sort_stories_by_property_and_order,
    sort_stories_by_raise_ranking,
)

__all__ = ["run_steps_and_sort_excel_file", "generate_jira_field_mapping_file"]

# Currently, the openpyxl package will report an obsolete warning.
warnings.simplefilter(action="ignore", category=UserWarning)
# Disable the HTTPS certificate verification warning.
disable_warnings()

HERE = pathlib.Path(__file__).resolve().parent
ASSETS = HERE / "assets"


def _get_jira_client() -> Optional[JiraClient]:
    load_dotenv(ASSETS / ".env")

    jira_url: Optional[str] = environ.get("JIRA_URL", default=None)
    if jira_url is None or jira_url.isspace() or len(jira_url) == 0:
        print(
            "The jira url is invalid. Please use the update-jira-info command to add/update url."
        )
        return None

    jira_acccess_token: Optional[str] = environ.get("JIRA_ACCESS_TOKEN", default=None)
    if (
        jira_acccess_token is None
        or jira_acccess_token.isspace()
        or len(jira_acccess_token) == 0
    ):
        print(
            "The jira access token is invalid. Please use the update-jira-info command to add/update token."
        )
        return None

    jira_client = JiraClient(jira_url, jira_acccess_token)

    if not jira_client.health_check():
        print(
            "The jira access token is revoked. Please use the update-jira-info command to add/update token."
        )
        return None

    print(f"Jira info: {jira_url}")

    return jira_client


def _query_jira_information(
    stories: List[Story], excel_definition: ExcelDefinition
) -> bool:
    jira_client = _get_jira_client()

    if jira_client is None:
        return False

    jira_fields = []

    for definition_column in excel_definition.get_columns():
        if (
            definition_column["name"] is None
            or definition_column["jira_field_mapping"] is None
        ):
            continue
        jira_fields.append(
            {
                "name": definition_column["name"],
                "jira_name": definition_column["jira_field_mapping"]["name"],
                "jira_path": definition_column["jira_field_mapping"]["path"],
            }
        )

    jira_query_result = jira_client.get_stories_detail(
        [story["storyId"].strip() for story in stories], jira_fields
    )

    for story in stories:
        story_id: str = story["storyId"].lower().strip()
        if story_id in jira_query_result:
            for jira_field in jira_fields:
                story[jira_field["name"]] = jira_query_result[story_id][
                    jira_field["jira_name"]
                ]
        else:
            # Story ID has been changed because of convertion.
            temp_result = jira_client.get_stories_detail([story_id], jira_fields)
            if len(temp_result) > 0:
                story["storyId"] = list(temp_result.keys())[0].upper()
                for jira_field in jira_fields:
                    story[jira_field["name"]] = list(temp_result.values())[0][
                        jira_field["jira_name"]
                    ].upper()
                print(
                    f"Story id has been changed. Previous: {story_id.upper()}, Current: {story['storyId'].upper()}"
                )
            else:
                print(f"Cannot find related information for story: {story_id}")
                story.need_sort = False
                continue

    return True


def run_steps_and_sort_excel_file(
    input_file: Union[str, Path],
    output_file: Union[str, Path],
    excel_definition_file: Optional[Union[str, Path]] = None,
    sprint_schedule_file: Optional[Union[str, Path]] = None,
    over_write: bool = True,
):
    """
    Sort the excel file and output the result

    :parm input_file:
        The excel file need to be sorted. (Absolute path only)

    :parm output_file:
        The sorted excel file location. (Absolute path only)

    :parm sprint_schedule_file:
        The JSON file which contains the priority list to calculate the :py:class:`Milestone`

    :parm excel_definition_file:
        The JSON file which contains the input excel file's structure.

    :parm over_write:
        Whether or not the exist output file will be over-write.
    """
    sprint_schedule = SprintScheduleStore()
    if sprint_schedule_file is None:
        print("Using default sprint schedule...")
        sprint_schedule.load(
            files("jira_assistant.assets")
            .joinpath("sprint_schedule.json")
            .read_text(encoding="utf-8")
        )
    else:
        print("Using custom sprint schedule...")
        sprint_schedule.load_file(sprint_schedule_file)

    excel_definition = ExcelDefinition()
    if excel_definition_file is None:
        print("Using default excel definition...")
        excel_definition.load(
            files("jira_assistant.assets")
            .joinpath("excel_definition.json")
            .read_text(encoding="utf-8")
        )
    else:
        print("Using custom excel definition...")
        excel_definition.load_file(excel_definition_file)

    validation_result = excel_definition.validate()
    if len(validation_result) != 0:
        print(
            "Validating excel definition failed. Please check below information to fix first."
        )
        for item in validation_result:
            print(item)
        return
    print("Validating excel definition success.")

    excel_columns, stories = read_excel_file(
        input_file, excel_definition, sprint_schedule
    )

    if stories is None or len(stories) == 0:
        print("There are no stories inside the excel file.")
        return

    # Execute pre-process steps
    pre_process_steps = excel_definition.get_pre_process_steps()

    for pre_process_step in pre_process_steps:
        print(f"Executing step: {pre_process_step['name']}...")
        if pre_process_step["name"].lower() in "RetrieveJiraInformation".lower():
            need_call_jira_api: bool = False
            for excel_definition_column in excel_definition.get_columns():
                if excel_definition_column["jira_field_mapping"] is not None:
                    need_call_jira_api = True
                    break

            if need_call_jira_api:
                stories_need_call_jira: List[Story] = []
                for story in stories:
                    if story.need_sort:
                        stories_need_call_jira.append(story)
                if not _query_jira_information(
                    stories_need_call_jira, excel_definition
                ):
                    print("Retrieve jira information failed.")
                    return
        elif pre_process_step["name"].lower() in "FilterOutStoryWithoutId".lower():
            for story in stories:
                if story["storyId"] is None:
                    story.need_sort = False
        elif (
            pre_process_step["name"].lower()
            in "FilterOutStoryBasedOnJiraStatus".lower()
        ):
            for story in stories:
                if story["status"] is not None and story[
                    "status"
                ].upper() in pre_process_step["config"].get("JiraStatuses", []):
                    story.need_sort = False
        print("Executing finish.")

    stories_no_need_sort = []
    stories_need_sort = []

    for story in stories:
        if story.need_sort:
            stories_need_sort.append(story)
        else:
            stories_no_need_sort.append(story)

    # Execute sorting logic.
    sort_strategies = excel_definition.get_sort_strategies()

    for sort_strategy in sort_strategies:
        print(f"Executing {sort_strategy['name']} sorting...")
        if sort_strategy["name"] is None:
            continue
        if sort_strategy["name"].lower() in "InlineWeights".lower():
            stories_need_sort = sort_stories_by_inline_weights(stories_need_sort)
        elif sort_strategy["name"].lower() in "SortOrder".lower():
            sort_stories_by_property_and_order(
                stories_need_sort, excel_definition, sort_strategy["config"]
            )
        elif sort_strategy["name"].lower() in "RaiseRanking".lower():
            stories_need_sort = sort_stories_by_raise_ranking(
                stories_need_sort, excel_definition, sort_strategy["config"]
            )
        print("Executing finish.")

    output_to_excel_file(
        output_file,
        stories_need_sort + stories_no_need_sort,  # First output the sorted stories.
        excel_definition,
        excel_columns,
        over_write,
    )

    print(f"{output_file} has been saved.")


def generate_jira_field_mapping_file(
    file: Union[str, Path], over_write: bool = True
) -> bool:
    jira_client = _get_jira_client()

    if jira_client is None:
        return False

    output_file_path: Path = Path(file).absolute()

    if output_file_path.exists():
        if over_write:
            try:
                remove(file)
            except PermissionError as e:
                raise FileExistsError(
                    f"The exist jira field mapping file: {file} cannot be removed. {e.args[0]}"
                ) from e
        else:
            raise FileExistsError(
                f"The jira field mapping file: {file} is already exist."
            )

    with open(output_file_path, mode="x", encoding="utf-8") as output_file:
        dump(jira_client.get_all_fields(), output_file)
        output_file.flush()
        output_file.close()

    return True



# -*- coding: utf-8 -*-
"""
This module offers a set of operations that user can modify their excel files.
"""
import pathlib
import warnings
from os import remove
from pathlib import Path
from typing import List, Optional, Tuple, Union

import openpyxl
from openpyxl.workbook import Workbook
from openpyxl.worksheet._read_only import ReadOnlyWorksheet
from openpyxl.worksheet.worksheet import Worksheet
from urllib3 import disable_warnings

from .excel_definition import ExcelDefinition
from .sprint_schedule import SprintScheduleStore
from .story import Story, StoryFactory
from .utils import is_absolute_path_valid

__all__ = ["read_excel_file", "output_to_excel_file"]

# Currently, the openpyxl package will report an obsolete warning.
warnings.simplefilter(action="ignore", category=UserWarning)
# Disable the HTTPS certificate verification warning.
disable_warnings()

HERE = pathlib.Path(__file__).resolve().parent
ASSETS = HERE / "assets"


def read_excel_file(
    file: Union[str, Path],
    excel_definition: ExcelDefinition,
    sprint_schedule: SprintScheduleStore,
) -> Tuple[List[str], List[Story]]:
    """
    Read and parse the excel file

    :parm file:
        The excel file that you want to read

    :parm excel_definition:
        The excel column definition which is imported
        from the :py:class:`ExcelDefinition`

    :parm sprint_schedule:
        The priority mapping for the :py:class:`Milestone` object.

    :return:
        A :py:class:`tuple` object which contains a list of column
        name and a list of :py:class:`Story`.
    """
    if not is_absolute_path_valid(file):
        raise FileNotFoundError(
            f"Please make sure the input excel file exist and the path should be absolute. File: {file}."
        )

    with open(str(file), mode="rb") as raw_file:
        work_book: Workbook = openpyxl.load_workbook(
            raw_file, read_only=True, keep_vba=False, data_only=True, keep_links=True
        )

        if work_book.active is None or (
            not isinstance(work_book.active, Worksheet)
            and not isinstance(work_book.active, ReadOnlyWorksheet)
        ):
            work_book.close()
            raise ValueError("The input excel file doesn't contain any sheets.")

        sheet: Union[Worksheet, ReadOnlyWorksheet] = work_book.active

        # The count of column is taking from the definition file to avoid too
        # many columns inside the excel file. Also, need to avoid exceed
        # the range of the actual count.
        column_count = min(excel_definition.max_column_index, sheet.max_column)

        if sheet.max_row < 2:
            work_book.close()
            return ([], [])

        columns: List[str] = []

        for column_index in range(1, column_count + 1):
            columns.append(str(sheet.cell(row=1, column=column_index).value))

        stories = []

        excel_definition_columns = excel_definition.get_columns()
        story_factory = StoryFactory(excel_definition_columns)

        for row in sheet.iter_rows(
            min_row=2, max_row=sheet.max_row, min_col=1, max_col=len(columns)
        ):
            if _should_skip(row):
                continue

            story: Story = story_factory.create_story()
            for column_index, column in enumerate(row):
                definition_column = excel_definition_columns[column_index]
                if definition_column["name"] is None:
                    continue
                story.set_value(
                    definition_column["type"], definition_column["name"], column.value
                )

            story.calc_sprint_schedule(sprint_schedule)
            stories.append(story)

        work_book.close()
        raw_file.close()
    return (columns, stories)


def _should_skip(row: tuple) -> bool:
    if len(row) == 0:
        return True
    first_cell_value = row[0].value
    if first_cell_value is None or len(str(first_cell_value)) == 0:
        return True
    return False


def output_to_excel_file(
    file: Union[str, Path],
    stories: "List[Story]",
    excel_definition: ExcelDefinition,
    columns_in_excel: Optional[List[str]] = None,
    over_write: bool = True,
):
    """
    Generate excel file

    :parm file:
        Output excel file name including the path

    :parm stories:
        A list of :py:class:`Story` which need to be wrote to the excel

    :parm excel_definition:
        The excel column definition which is imported from
        the :py:class:`ExcelDefinition`

    :parm columns_in_excel:
        Using separate column names instead of importing from
        the :py:class:`ExcelDefinition`. Usually, it comes from the
        input excel file.

    :parm over_write:
        Whether or not the exist output file will be over-write.
    """
    if file is None or not pathlib.Path(file).is_absolute():
        raise ValueError("The output file path is invalid.")

    if pathlib.Path(file).exists():
        if over_write is True:
            try:
                remove(file)
            except PermissionError as e:
                raise FileExistsError(
                    f"The exist excel file: {file} cannot be removed. {e.args[0]}"
                ) from e
        else:
            raise FileExistsError(f"The output excel file: {file} is already exist.")

    work_book = openpyxl.Workbook(write_only=False)

    if work_book.active is None or (
        not isinstance(work_book.active, Worksheet)
        and not isinstance(work_book.active, Worksheet)
    ):
        work_book.close()
        raise ValueError("The output excel file cannot be generated.")

    sheet: Worksheet = work_book.active

    excel_definition_columns = excel_definition.get_columns()

    # Use original excel column name first.
    columns = columns_in_excel
    if columns is None:
        columns = [column["name"] for column in excel_definition_columns]

    for column_index, column in enumerate(columns):
        cell = sheet.cell(row=1, column=column_index + 1)
        # There are three kinds of Cells. Only the Cell has the value attribute.
        if hasattr(cell, "value"):
            setattr(cell, "value", column)

    if stories is not None and stories:
        for row_index, story in enumerate(stories):
            for column in excel_definition_columns:
                if column["name"] is None:
                    continue
                cell = sheet.cell(row=row_index + 2, column=column["index"])
                if hasattr(cell, "value"):
                    setattr(cell, "value", story.format_value(column["name"]))

    work_book.save(str(file))
    work_book.close()



# -*- coding: utf-8 -*-
"""
This module is used to store excel column definition information.
"""
import warnings
from sys import version_info
from typing import Any, Dict, List, Optional, TypedDict, Union

from jira import JIRA, JIRAError
from urllib3 import disable_warnings

from .story import Story
from .utils import JiraFieldTypeDefinition, get_jira_field_type

if version_info < (3, 11):
    from typing_extensions import NotRequired
else:
    from typing import NotRequired

# Currently, the openpyxl package will report an obsolete warning.
warnings.simplefilter(action="ignore", category=UserWarning)
# Disable the HTTPS certificate verification warning.
disable_warnings()


def connect_field_path(path_a: str, path_b: str) -> str:
    return path_a + "." + path_b


class JiraClient:
    def __init__(self, url: str, access_token: str) -> None:
        self.jira = JIRA(
            server=url,
            token_auth=access_token,
            timeout=20,
            options={"verify": False},
        )
        self._field_cache: Dict[str, Dict[str, Union[str, List[str], bool]]] = {}

    def health_check(self) -> bool:
        try:
            if self.jira.myself() is not None:
                return True
            return False
        except JIRAError:
            return False

    def create_storys(self, storys: List[Story]) -> "List[Story]":
        # self.jira.create_issues(, prefetch=true)
        return storys

    def get_all_fields(self) -> "Dict[str, Dict[str, Union[str, List[str], bool]]]":
        if not self._field_cache:
            for field in self.jira.fields():
                if "schema" not in field.keys():
                    continue

                temp: Dict[str, Union[str, List[str], bool]] = {
                    "id": field["id"],
                    "isArray": False,
                }

                class FieldSchema(TypedDict):
                    type: str
                    items: NotRequired[str]

                schema: FieldSchema = field["schema"]

                if schema["type"] == "array":
                    temp["isArray"] = True
                    field_type_name = schema.get("items", None)
                else:
                    field_type_name = schema["type"]

                field_type: Optional[JiraFieldTypeDefinition] = get_jira_field_type(
                    field_type_name
                )

                path_list: List[str] = []
                if field_type is not None and field_type.get("isBasic", None) is False:
                    self._append_list(
                        path_list,
                        self._query_all_properties(field_type, field["id"]),
                    )
                else:
                    # Usually, the format is like below:
                    # "customfield_11005": {
                    #     "self": "https://jira_server/rest/api/2/customFieldOption/11005",
                    #     "value": "No",
                    #     "id": "11005",
                    #     "disabled": false
                    # },
                    path_list.append(connect_field_path(field["id"], "id"))
                    path_list.append(connect_field_path(field["id"], "value"))

                temp["path"] = path_list
                self._field_cache[field["name"]] = temp
        return self._field_cache

    def _query_all_properties(
        self,
        field_type: Optional[JiraFieldTypeDefinition],
        root_path: str,
    ) -> List[str]:
        path_list: List[str] = []

        if field_type is None:
            return path_list

        if field_type.get("isBasic", None) is True:
            self._append_str_to_list(path_list, field_type.get("name", ""))
        elif field_type.get("isArray", None) is True and "itemType" in field_type:
            item_field_type = get_jira_field_type(field_type["itemType"])
            if item_field_type is not None:
                self._append_list(
                    path_list,
                    self._query_all_properties(
                        item_field_type, item_field_type.get("name", "")
                    ),
                )
        elif "properties" in field_type:
            for child_field_type in field_type["properties"]:
                if (
                    child_field_type.get("isArray", None) is True
                    and "itemType" in child_field_type
                ):
                    child_field = get_jira_field_type(child_field_type["itemType"])
                else:
                    child_field = get_jira_field_type(child_field_type["type"])

                if child_field is not None:
                    new_root_path = connect_field_path(
                        root_path, child_field_type.get("name", "")
                    )

                    if child_field.get("isBasic") is True:
                        path_list.append(new_root_path)
                    elif (
                        child_field.get("isArray", None) is True
                        and "itemType" in child_field
                    ):
                        self._append_list(
                            path_list,
                            self._query_all_properties(
                                get_jira_field_type(child_field["itemType"]),
                                new_root_path,
                            ),
                        )
                    else:
                        self._append_list(
                            path_list,
                            self._query_all_properties(child_field, new_root_path),
                        )
        else:
            pass

        return path_list

    def _append_list(self, arr_a: List[str], arr_b: List[str]):
        for item in arr_b:
            arr_a.append(item)

    def _append_str_to_list(self, arr: List[str], statement: str):
        for i, _ in enumerate(arr):
            arr[i] = connect_field_path(arr[i], statement)

    def get_stories_detail(
        self, story_ids: List[str], jira_fields: List[Dict[str, str]]
    ) -> "Dict[str, Dict[str, str]]":
        final_result = {}
        batch_size = 200

        try:
            if len(story_ids) > batch_size:
                start_index = 0
                end_index = batch_size
                while end_index <= len(story_ids) and start_index < len(story_ids):
                    # print(f"Start: {start_index}, End: {end_index}")
                    final_result.update(
                        self._internal_get_stories_detail(
                            story_ids[start_index:end_index], jira_fields
                        )
                    )
                    start_index = end_index
                    if start_index + batch_size < len(story_ids):
                        end_index = start_index + batch_size
                    else:
                        end_index = start_index + (len(story_ids) - end_index)
                return final_result

            return self._internal_get_stories_detail(story_ids, jira_fields)
        except JIRAError as e:
            print(
                f"Calling JIRA API failed. HttpStatusCode: {e.status_code}. Response: {e.response.json()}"
            )

            return {}

    def _internal_get_stories_detail(
        self, story_ids: List[str], jira_fields: List[Dict[str, str]]
    ) -> "Dict[str, Dict[str, str]]":
        id_query = ",".join([f"'{str(story_id)}'" for story_id in story_ids])

        try:
            search_result: Dict[str, Any] = self.jira.search_issues(
                jql_str=f"id in ({id_query})",
                maxResults=len(story_ids),
                fields=[field["jira_name"] for field in jira_fields],
                json_result=True,
            )  # type: ignore

            final_result = {}
            for issue in search_result["issues"]:
                fields_result = {}
                for field in jira_fields:
                    # First element in the tuple is jira field name like "customfield_13210 or status..."
                    field_name = field["jira_name"]
                    # Remain elements represent the property path.
                    field_value: Any = issue["fields"]
                    for field_path in field["jira_path"].split("."):
                        if field_value is None:
                            field_value = ""
                            break
                        field_value = field_value.get(field_path, None)
                    fields_result[field_name] = field_value
                final_result[issue["key"].lower()] = fields_result

            return final_result
        except JIRAError as e:
            print(
                f"Calling JIRA API failed. HttpStatusCode: {e.status_code}. Response: {e.response.json()}"
            )

            return {}



# -*- coding: utf-8 -*-
"""
This module is used to list all exported classes/methods.
"""
try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

from .assistant import run_steps_and_sort_excel_file
from .console_script import generate_template, process_excel_file, update_jira_info
from .excel_definition import ExcelDefinition, ExcelDefinitionColumn
from .excel_operation import output_to_excel_file, read_excel_file
from .milestone import Milestone
from .priority import Priority
from .sprint_schedule import SprintScheduleStore
from .story import Story, StoryFactory

__version__ = metadata.version("jira_assistant")

__all__ = [
    "ExcelDefinition",
    "ExcelDefinitionColumn",
    "read_excel_file",
    "output_to_excel_file",
    "run_steps_and_sort_excel_file",
    "Milestone",
    "Priority",
    "SprintScheduleStore",
    "Story",
    "StoryFactory",
    "process_excel_file",
    "generate_template",
    "update_jira_info",
]



"""
Milestone: 2 weeks                                           2022/1/10
Milestone Start Date
Release: 5 weeks (S1 2 weeks S2 2 weeks Harden 1 week)       2022/1/10
S1 Start Date 
S2 Start Date
"""


from typing import Any

from .sprint_schedule import SprintScheduleStore

__all__ = ["Milestone"]


class Milestone:
    # __init__ method cannot have classmethod attribute! Otherwise, all instance will point to the same one.
    def __init__(self, raw: Any) -> None:
        self.raw: str = raw
        self.sprint = self.raw
        self.priority = 0

    @property
    def priority(self) -> int:
        return self._priority

    @priority.setter
    def priority(self, value: int):
        self._priority = value

    def calc_priority(self, sprint_schedule: SprintScheduleStore) -> None:
        if self.sprint is None:
            self.priority = 0
        else:
            self.priority = sprint_schedule.get_priority(self.sprint)

    def __str__(self) -> str:
        if self.raw is None:
            return ""
        return self.raw

    def __lt__(self, __o: "Milestone") -> bool:
        if self.priority < __o.priority:
            return True
        return False

    def __gt__(self, __o: "Milestone") -> bool:
        if self.priority > __o.priority:
            return True
        return False

    def __le__(self, __o: "Milestone") -> bool:
        if self.priority <= __o.priority:
            return True
        return False

    def __ge__(self, __o: "Milestone") -> bool:
        if self.priority >= __o.priority:
            return True
        return False

    def __eq__(self, __o: "Milestone") -> bool:
        if self.priority == __o.priority:
            return True
        return False



import pathlib

from pytest import raises

from jira_assistant.milestone import Milestone
from jira_assistant.sprint_schedule import SprintScheduleStore

HERE = pathlib.Path(__file__).resolve().parent
SRC_ASSETS = HERE.parent / "src/jira_assistant/assets"


class TestMilestone:
    def test_init(self):
        schedule_filename = SRC_ASSETS / "sprint_schedule.json"
        store = SprintScheduleStore()
        store.load_file(schedule_filename)

        milestone = Milestone("R139")
        milestone.calc_priority(store)
        assert milestone.priority == 1

    def test_compare(self):
        schedule_filename = SRC_ASSETS / "sprint_schedule.json"
        store = SprintScheduleStore()
        store.load_file(schedule_filename)

        m_1 = Milestone("M123")
        m_1.calc_priority(store)
        m_2 = Milestone("M125")
        m_2.calc_priority(store)
        m_3 = Milestone("m125")
        m_3.calc_priority(store)
        m_4 = Milestone("R141")
        m_4.calc_priority(store)
        assert m_1 < m_2
        with raises(AssertionError):
            assert m_1 > m_2
        assert m_1 <= m_2
        with raises(AssertionError):
            assert m_1 >= m_2
        assert m_2 >= m_1
        assert m_2 == m_3
        assert m_3 == m_4
        assert m_4 > m_1
        with raises(AssertionError):
            assert m_1 == m_4

    def test_str(self):
        m_1 = Milestone(None)

        assert len(str(m_1)) == 0



from jira_assistant.priority import Priority


class TestPriority:
    def test_compare_priority(self):
        p_1: Priority = Priority.CRITICAL
        p_2: Priority = Priority.HIGH
        p_3: Priority = Priority.HIGH
        p_4: Priority = Priority.MIDDLE
        p_5: Priority = Priority.LOW
        p_6: Priority = Priority.NA

        assert p_1 > p_2
        assert p_2 < p_1
        assert p_2 == p_3
        assert p_3 > p_4
        assert p_4 < p_3
        assert p_4 > p_5
        assert p_5 < p_4
        assert p_5 > p_6
        assert p_5 >= p_6
        assert p_6 < p_5
        assert p_6 <= p_5

    def test_priority_to_str(self):
        p_1: Priority = Priority.CRITICAL
        p_2: Priority = Priority.NA
        assert str(p_1) == "Critical"
        assert str(p_2) == "N/A"



from __future__ import annotations

import os

from mock_server import (
    mock_jira_requests,
    mock_jira_requests_with_failed_status_code,
    mock_jira_stories,
)
from requests_mock import Mocker

from jira_assistant.jira_client import JiraClient


class TestJiraClient:
    def test_get_stories_detail(self):
        with Mocker(
            real_http=False, case_sensitive=False, adapter=mock_jira_requests()
        ):
            client = JiraClient(os.environ["JIRA_URL"], os.environ["JIRA_ACCESS_TOKEN"])

            stories = client.get_stories_detail(
                ["A-1", "A-2", "B-1"],
                [
                    {
                        "name": "domain",
                        "jira_name": "customfield_15601",
                        "jira_path": "customfield_15601.value",
                    },
                    {
                        "name": "status",
                        "jira_name": "status",
                        "jira_path": "status.name",
                    },
                ],
            )
            assert len(stories) == 3

    def test_get_stories_detail_with_large_amount_of_stories(self):
        with Mocker(
            real_http=False, case_sensitive=False, adapter=mock_jira_requests()
        ):
            client = JiraClient(os.environ["JIRA_URL"], os.environ["JIRA_ACCESS_TOKEN"])

            stories = client.get_stories_detail(
                list(mock_jira_stories.keys()),
                [
                    {
                        "name": "domain",
                        "jira_name": "customfield_15601",
                        "jira_path": "customfield_15601.value",
                    },
                    {
                        "name": "status",
                        "jira_name": "status",
                        "jira_path": "status.name",
                    },
                ],
            )
            assert len(stories) == 246

    def test_health_check(self):
        with Mocker(real_http=False, adapter=mock_jira_requests()):
            client = JiraClient(os.environ["JIRA_URL"], os.environ["JIRA_ACCESS_TOKEN"])

            assert client.health_check() is True

    def test_health_check_failed(self):
        with Mocker(
            real_http=False, adapter=mock_jira_requests_with_failed_status_code()
        ):
            client = JiraClient(os.environ["JIRA_URL"], os.environ["JIRA_ACCESS_TOKEN"])

            assert client.health_check() is False

    def test_get_stories_detail_failed(self):
        with Mocker(
            real_http=False,
            case_sensitive=False,
            adapter=mock_jira_requests_with_failed_status_code(),
        ):
            client = JiraClient(os.environ["JIRA_URL"], os.environ["JIRA_ACCESS_TOKEN"])

            stories = client.get_stories_detail(
                ["A-1", "A-2", "B-1"],
                [
                    {
                        "name": "domain",
                        "jira_name": "customfield_15601",
                        "jira_path": "customfield_15601.value",
                    },
                    {
                        "name": "status",
                        "jira_name": "status",
                        "jira_path": "status.name",
                    },
                ],
            )
            assert len(stories) == 0

    def test_get_stories_detail_with_large_amount_of_stories_failed(self):
        with Mocker(
            real_http=False,
            case_sensitive=False,
            adapter=mock_jira_requests_with_failed_status_code(),
        ):
            client = JiraClient(os.environ["JIRA_URL"], os.environ["JIRA_ACCESS_TOKEN"])

            stories = client.get_stories_detail(
                list(mock_jira_stories.keys()),
                [
                    {
                        "name": "domain",
                        "jira_name": "customfield_15601",
                        "jira_path": "customfield_15601.value",
                    },
                    {
                        "name": "status",
                        "jira_name": "status",
                        "jira_path": "status.name",
                    },
                ],
            )
            assert len(stories) == 0

    def test_get_all_fields(self):
        with Mocker(
            real_http=False,
            case_sensitive=False,
            adapter=mock_jira_requests(),
        ):
            client = JiraClient(os.environ["JIRA_URL"], os.environ["JIRA_ACCESS_TOKEN"])

            result = client.get_all_fields()

            assert len(result) == 5



import re
from datetime import datetime
from decimal import Decimal
from functools import cmp_to_key
from operator import attrgetter
from typing import Any, Dict, List, Optional, Set, Tuple

from dateutil import parser

from .excel_definition import ExcelDefinition, ExcelDefinitionColumn
from .milestone import Milestone
from .priority import Priority, convert_to_priority
from .sprint_schedule import SprintScheduleStore

__all__ = [
    "Story",
    "StoryFactory",
    "convert_to_bool",
    "convert_to_datetime",
    "convert_to_decimal",
    "sort_stories_by_property_and_order",
    "sort_stories_by_raise_ranking",
    "sort_stories_by_inline_weights",
    "compare_story_based_on_inline_weights",
]


def convert_to_bool(raw: Any) -> bool:
    if isinstance(raw, bool):
        return raw
    value = str(raw).strip().upper()
    if value in ("YES", "TRUE"):
        return True
    return False


def convert_to_decimal(raw: Any) -> Decimal:
    if isinstance(raw, Decimal):
        return raw
    raw = str(raw).strip()
    pattern = re.compile("[0-9.]{1,10}")
    result = pattern.search(raw)
    if result is not None:
        return Decimal(result.group())
    return Decimal(0)


def convert_to_datetime(raw: Any) -> Optional[datetime]:
    if isinstance(raw, datetime):
        return raw
    if raw is None:
        return None
    raw = str(raw).strip()
    return parser.parse(raw)


class Story:
    def __init__(self, factory: "StoryFactory") -> None:
        self.need_sort = True
        if factory is None:
            raise ValueError("Story must be created from a specific factory!")
        self.factory = factory
        for column in self.factory.columns:
            if column["name"] is None:
                continue
            if column["type"] is str:
                setattr(self, column["name"], "")
            elif column["type"] is bool:
                setattr(self, column["name"], False)
            elif column["type"] is Priority:
                setattr(self, column["name"], Priority.NA)
            elif column["type"] is Milestone:
                setattr(self, column["name"], None)
            elif column["type"] is datetime:
                setattr(self, column["name"], None)
            else:
                setattr(self, column["name"], None)

    @property
    def need_sort(self) -> bool:
        return self._need_sort

    @need_sort.setter
    def need_sort(self, value: bool):
        self._need_sort = value

    def __getitem__(self, property_name) -> Any:
        return getattr(self, property_name)

    def format_value(self, property_name: str) -> str:
        property_value = getattr(self, property_name, None)
        if property_value is None:
            return ""
        if isinstance(property_value, datetime):
            return property_value.date().isoformat()
        if isinstance(property_value, bool):
            if property_value:
                return "Yes"
            return "No"
        if isinstance(property_value, float):
            return str(property_value)
        return str(property_value)

    def set_value(self, property_type: Any, property_name: str, property_value: Any):
        if property_type is str:
            setattr(self, property_name, property_value)
        elif property_type is bool:
            setattr(self, property_name, convert_to_bool(property_value))
        elif property_type is Priority:
            setattr(self, property_name, convert_to_priority(property_value))
        elif property_type is datetime:
            setattr(self, property_name, convert_to_datetime(property_value))
        elif property_type is Milestone:
            milestone = Milestone(property_value)
            setattr(self, property_name, milestone)
        else:
            setattr(self, property_name, property_value)

    def __setitem__(self, property_name, property_value):
        self.set_value(type(property_value), property_name, property_value)

    def calc_sprint_schedule(self, sprint_schedule: SprintScheduleStore):
        for column in self.factory.columns:
            if column["type"] is Milestone and self[column["name"]] is not None:
                self[column["name"]].calc_priority(sprint_schedule)

    # Currently, comparing story only consider the Priority properties.
    def __lt__(self, __o: "Optional[Story]") -> bool:
        raise TypeError("unsupported operand type(s) for comparing story.")

    def __le__(self, __o: "Optional[Story]") -> bool:
        raise TypeError("unsupported operand type(s) for comparing story.")

    def __gt__(self, __o: "Optional[Story]") -> bool:
        raise TypeError("unsupported operand type(s) for comparing story.")

    def __ge__(self, __o: "Optional[Story]") -> bool:
        raise TypeError("unsupported operand type(s) for comparing story.")

    def __eq__(self, __o: "Optional[Story]") -> bool:
        raise TypeError("unsupported operand type(s) for comparing story.")

    # For CSV consideration.
    def __str__(self):
        result = ""
        if self is None:
            return result

        separator = ", "
        for column in self.factory.columns:
            if column["name"] is not None and hasattr(self, column["name"]):
                result += f"{str(getattr(self, column['name']))}{separator}"
        return result


class StoryFactory:
    def __init__(self, columns: "List[ExcelDefinitionColumn]") -> None:
        if columns is None:
            raise ValueError("Columns must be provided!")
        self._columns = columns
        self._inline_weight_compare_rules = (
            self.__generate_inline_weights_compare_rules()
        )

    def __generate_inline_weights_compare_rules(self) -> "List[Tuple[str, int]]":
        rules = []
        for column in self._columns:
            if column["inline_weights"] > 0:
                rules.append((column["name"], column["inline_weights"]))
        rules.sort(key=lambda r: r[1], reverse=True)
        return rules

    @property
    def columns(self):
        return self._columns

    @property
    def inline_weights_compare_rules(self) -> "List[Tuple[str, int]]":
        return self._inline_weight_compare_rules

    def create_story(self) -> Story:
        return Story(self)


def sort_stories_by_inline_weights(stories: "List[Story]") -> "List[Story]":
    return sorted(
        stories, key=cmp_to_key(compare_story_based_on_inline_weights), reverse=True
    )


def compare_story_based_on_inline_weights(
    story_a: Optional[Story], story_b: Optional[Story]
) -> int:
    """
    Compare two stories.

    :parm a:
        First story
    :parm b:
        Second story
    :parm sort_rule:
        Priority information
    :return
        1: means a > b
        0: means a == b
        -1: means a < b
    """
    if story_a is None or story_b is None:
        raise ValueError("The compare stories cannot be None.")

    if (
        story_a.factory != story_b.factory
        or story_a.factory is None
        or story_b.factory is None
    ):
        raise ValueError("The compare stories were built by different factory.")

    # story a and b have the same factory.
    compare_rules: List[Tuple[str, int]] = story_a.factory.inline_weights_compare_rules

    rules_count = len(compare_rules)

    if rules_count == 0:
        return 0

    skip_index_of_a = []
    skip_index_of_b = []
    count = rules_count
    while count > 0:
        # property_value, property_location
        highest_property_of_a = None
        highest_property_of_b = None
        for i, compare_rule in enumerate(compare_rules):
            if i in skip_index_of_a:
                continue

            if highest_property_of_a is None:
                # property_value, property_location
                highest_property_of_a = (story_a[compare_rule[0]], i)

            if story_a[compare_rule[0]] > highest_property_of_a[0]:
                highest_property_of_a = (story_a[compare_rule[0]], i)

        for i, compare_rule in enumerate(compare_rules):
            if i in skip_index_of_b:
                continue

            if highest_property_of_b is None:
                highest_property_of_b = (story_b[compare_rule[0]], i)

            if story_b[compare_rule[0]] > highest_property_of_b[0]:
                highest_property_of_b = (story_b[compare_rule[0]], i)

        if highest_property_of_a is None:
            highest_property_of_a = (Priority.NA, count)
        else:
            skip_index_of_a.append(highest_property_of_a[1])

        if highest_property_of_b is None:
            highest_property_of_b = (Priority.NA, count)
        else:
            skip_index_of_b.append(highest_property_of_b[1])

        # priority value
        if highest_property_of_a[0] > highest_property_of_b[0]:
            return 1
        if highest_property_of_a[0] == highest_property_of_b[0]:
            if highest_property_of_a[1] < highest_property_of_b[1]:
                return 1
            if highest_property_of_a[1] > highest_property_of_b[1]:
                return -1
        else:
            return -1

        # property location
        if highest_property_of_a[1] > highest_property_of_b[1]:
            return 1
        if highest_property_of_a[1] == highest_property_of_b[1]:
            if highest_property_of_a[0] > highest_property_of_b[0]:
                return 1
            if highest_property_of_a[0] < highest_property_of_b[0]:
                return -1
        else:
            return -1

        count -= 1
        continue
    return 0


def sort_stories_by_property_and_order(
    stories: "List[Story]", excel_definition: ExcelDefinition, config: dict
):
    sort_rules: List[Tuple[str, bool]] = []
    excel_definition_columns = excel_definition.get_columns()

    if "ParentScopeIndexRange" in config:
        column_definitions: Dict[int, ExcelDefinitionColumn] = {
            c["index"]: c for c in excel_definition_columns
        }

        for column in excel_definition_columns:
            if column["scope_require_sort"] is True and column["name"] is not None:
                sort_rules.append((column["name"], column["scope_sort_order"]))

        _internal_sort_stories_by_property_and_order_considering_parent_range(
            stories,
            column_definitions,
            sort_rules,
            config["ParentScopeIndexRange"],
        )
    else:
        for column in excel_definition_columns:
            if column["require_sort"] is True and column["name"] is not None:
                sort_rules.append((column["name"], column["sort_order"]))

        _internal_sort_stories_by_property_and_order(stories, sort_rules)


def _internal_sort_stories_by_property_and_order(
    stories: "List[Story]", sort_rules: "List[Tuple[str, bool]]"
):
    for column_name, sort_order in reversed(sort_rules):
        stories.sort(key=attrgetter(column_name), reverse=sort_order)


def _internal_sort_stories_by_property_and_order_considering_parent_range(
    stories: "List[Story]",
    story_columns: "Dict[int, ExcelDefinitionColumn]",
    sort_rules: "List[Tuple[str, bool]]",
    parent_level_index_range: "Set[int]",
) -> "List[Story]":
    begin_index = 0
    end_index = 0
    while end_index <= len(stories) - 1:
        for i in range(begin_index, len(stories) - 1):
            all_parent_column_matched = True
            for column_index in parent_level_index_range:
                if (
                    stories[i][story_columns[column_index]["name"]]
                    != stories[i + 1][story_columns[column_index]["name"]]
                ):
                    all_parent_column_matched = False
                    break
            if all_parent_column_matched:
                continue
            end_index = i
            break
        # Same parent level process
        for column_name, sort_order in reversed(sort_rules):
            if begin_index != end_index:
                stories[begin_index : end_index + 1] = sorted(
                    stories[begin_index : end_index + 1],
                    key=attrgetter(column_name),
                    reverse=sort_order,
                )

        begin_index = end_index + 1
        end_index = begin_index

    return stories


def sort_stories_by_raise_ranking(
    stories: "List[Story]", excel_definition: ExcelDefinition, config: Dict
) -> "List[Story]":
    if stories is None:
        return []
    sort_rules: List[Tuple[str, int]] = []
    excel_definition_columns = excel_definition.get_columns()

    result = []

    # New raise ranking mode.
    if "ParentScopeIndexRange" in config:
        for column in excel_definition_columns:
            if column["scope_raise_ranking"] > 0 and column["name"] is not None:
                sort_rules.append(
                    (
                        column["name"],
                        column["scope_raise_ranking"],
                    )
                )

        if len(sort_rules) == 0:
            return stories

        sort_rules.sort(key=lambda x: x[1], reverse=True)  # sort by scope_raise_ranking

        column_definitions: Dict[int, ExcelDefinitionColumn] = {
            c["index"]: c for c in excel_definition_columns
        }

        result = _internal_raise_story_ranking_by_property_considering_parent_level(
            stories,
            column_definitions,
            sort_rules,
            config["ParentScopeIndexRange"],
        )
    else:
        for column in excel_definition_columns:
            if column["raise_ranking"] > 0 and column["name"] is not None:
                sort_rules.append(
                    (
                        column["name"],
                        column["raise_ranking"],
                    )
                )

        if len(sort_rules) == 0:
            return stories

        sort_rules.sort(key=lambda x: x[1], reverse=True)  # sort by raise_ranking

        for property_name, _ in sort_rules:
            result = _internal_raise_story_ranking_by_property(stories, property_name)

    return result


def _internal_raise_story_ranking_by_property(
    stories: "List[Story]", property_name: str
) -> "List[Story]":
    if stories is None or len(stories) == 0:
        return stories
    # Use first story as example
    if not hasattr(stories[0], property_name):
        return stories
    return _raise_story_ranking_by_property(stories, property_name)


def _internal_raise_story_ranking_by_property_considering_parent_level(
    stories: "List[Story]",
    story_columns: "Dict[int, ExcelDefinitionColumn]",
    sort_rules: "List[Tuple[str, int]]",
    parent_level_index_range: "Set[int]",
) -> "List[Story]":
    begin_index = 0
    end_index = 0
    while end_index <= len(stories) - 1:
        for i in range(begin_index, len(stories) - 1):
            all_parent_column_matched = True
            for column_index in parent_level_index_range:
                if (
                    stories[i][story_columns[column_index]["name"]]
                    != stories[i + 1][story_columns[column_index]["name"]]
                ):
                    all_parent_column_matched = False
                    break
            if all_parent_column_matched:
                continue
            end_index = i
            break

        # Same parent level process
        for column_name, _ in reversed(sort_rules):
            stories[begin_index : end_index + 1] = _raise_story_ranking_by_property(
                stories[begin_index : end_index + 1], column_name
            )

        begin_index = end_index + 1
        end_index = begin_index

    return stories


# Only bool indicator for now
def _raise_story_ranking_by_property(
    stories: "List[Story]", property_name: str
) -> "List[Story]":
    if not isinstance(getattr(stories[0], property_name), bool):
        return stories
    result: List[Story] = [stories[0]] * len(stories)
    j = 0
    for story in stories:
        if getattr(story, property_name) is True:
            result[j] = story
            j += 1
    for story in stories:
        if getattr(story, property_name) is False:
            result[j] = story
            j += 1
    return result



# -*- coding: utf-8 -*-
"""
This module is used to provide the console program.
"""
import os
import pathlib
import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from datetime import datetime
from pathlib import Path
from shutil import copyfile
from typing import Optional
from urllib.parse import ParseResult, urlparse

from dotenv import set_key

from .assistant import generate_jira_field_mapping_file, run_steps_and_sort_excel_file
from .excel_definition import ExcelDefinition
from .excel_operation import output_to_excel_file

__all__ = ["process_excel_file", "generate_template", "update_jira_info"]


def get_args_for_process_excel_file() -> Namespace:
    parser = ArgumentParser(
        description="Jira tool: Used to pre-process and sort stories",
        formatter_class=ArgumentDefaultsHelpFormatter,
        allow_abbrev=False,
    )

    parser.add_argument(
        "input_file", metavar="input_file", type=pathlib.Path, help="Source Excel file."
    )
    parser.add_argument(
        "-o",
        "--output_folder",
        metavar="",
        type=pathlib.Path,
        required=False,
        help="Output folder.",
    )
    parser.add_argument(
        "--excel_definition_file",
        metavar="",
        type=pathlib.Path,
        required=False,
        help="Excel definition file. File format: JSON.",
    )
    parser.add_argument(
        "--sprint_schedule_file",
        metavar="",
        type=pathlib.Path,
        required=False,
        help="Milestone priority file. File format: JSON.",
    )
    parser.add_argument(
        "--over_write",
        metavar="",
        type=bool,
        required=False,
        help="Whether or not to over write existing file.",
    )

    args = parser.parse_args()

    return args


def process_excel_file() -> None:
    try:
        args = get_args_for_process_excel_file()

        # Pre-Process input file
        input_file_absolute_path: pathlib.Path = (
            pathlib.Path.cwd() / args.input_file.as_posix()
        ).resolve()

        if input_file_absolute_path.suffix.lower() != ".xlsx":
            print(f"Please provide an Excel file. File: {input_file_absolute_path}.")
            sys.exit(1)

        if not os.path.exists(input_file_absolute_path):
            print(f"Input file is not exist. File: {input_file_absolute_path}.")
            sys.exit(1)

        input_file_name_without_extension = input_file_absolute_path.stem

        # Pre-Process output file
        output_folder_absolute_path: pathlib.Path = (
            input_file_absolute_path.parent.absolute()
        )

        if args.output_folder is not None:
            output_folder_absolute_path = (
                pathlib.Path(args.output_folder).resolve().absolute()
            )

        if not output_folder_absolute_path.exists():
            output_folder_absolute_path.mkdir(parents=True, exist_ok=True)

        output_file_absolute_path: pathlib.Path = (
            output_folder_absolute_path
            / f"{input_file_name_without_extension}_sorted.xlsx"
        ).resolve()

        # Fix duplicate output file issue.
        test_output_file_absolute_path = pathlib.Path(
            output_file_absolute_path
        ).resolve()

        copy_count = 1
        while test_output_file_absolute_path.exists():
            test_output_file_absolute_path = (
                output_folder_absolute_path
                / f"{ output_file_absolute_path.stem }_{copy_count}.xlsx"
            )
            copy_count += 1

        output_file_absolute_path = pathlib.Path(
            test_output_file_absolute_path
        ).resolve()

        excel_definition_file_absolute_path = None

        if args.excel_definition_file is not None:
            excel_definition_file_absolute_path = pathlib.Path(
                pathlib.Path.cwd() / args.excel_definition_file.as_posix()
            ).resolve()

        sprint_schedule_file_absolute_path = None

        if args.sprint_schedule_file is not None:
            sprint_schedule_file_absolute_path = pathlib.Path(
                pathlib.Path.cwd() / args.sprint_schedule_file.as_posix()
            ).resolve()

        # Over write parameter.
        over_write = True
        if args.over_write is not None:
            over_write = args.over_write

        run_steps_and_sort_excel_file(
            input_file_absolute_path,
            output_file_absolute_path,
            excel_definition_file_absolute_path,
            sprint_schedule_file_absolute_path,
            over_write,
        )

        sys.exit(0)
    except Exception as e:
        print(e)
        sys.exit(1)


def get_args_for_generate_template() -> Namespace:
    parser = ArgumentParser(
        description="Jira tool: Used to generate templates",
        formatter_class=ArgumentDefaultsHelpFormatter,
        allow_abbrev=False,
    )

    parser.add_argument(
        "template_type",
        metavar="template_type",
        type=str,
        help="What kind of file template you want to generate. Choices: excel, excel-definition, sprint-schedule or jira-field-mapping.",
        choices=["excel", "excel-definition", "sprint-schedule", "jira-field-mapping"],
    )

    return parser.parse_args()


HERE = pathlib.Path(__file__).resolve().parent
SRC_ASSETS = HERE / "assets"


def generate_template():
    try:
        args = get_args_for_generate_template()

        template_type: str = str(args.template_type).lower()

        result: Optional[Path] = None
        if template_type == "excel":
            result = _generate_excel_template(
                _generate_timestamp_filename("excel-template", ".xlsx")
            )
        elif template_type == "excel-definition":
            result = copyfile(
                SRC_ASSETS / "excel_definition.json",
                _generate_timestamp_filename("excel-definition-template", ".json"),
            )
        elif template_type == "sprint-schedule":
            result = copyfile(
                SRC_ASSETS / "sprint_schedule.json",
                _generate_timestamp_filename("sprint-schedule-template", ".json"),
            )
        elif template_type == "jira-field-mapping":
            result = _generate_jira_field_mapping_template(
                _generate_timestamp_filename("jira-field-mapping", ".json")
            )
        else:
            print(
                "Invalid template type. Choices: excel, excel-definition or sprint-schedule."
            )

        if result is not None and result.is_file():
            print(f"Generate success! Template type: {template_type}.")
            sys.exit(0)
        else:
            print(f"Generate failed! Template type: {template_type}.")
            sys.exit(1)
    except Exception as e:
        print(e)
        sys.exit(1)


def _generate_timestamp_filename(prefix: str, extension: str) -> "Path":
    return (
        Path.cwd()
        / f'{prefix}-{datetime.now().strftime("%y-%m-%d-%H-%M-%S")}{extension}'
    ).resolve()


def _generate_excel_template(output_file: "Path") -> Optional[Path]:
    try:
        excel_definition = ExcelDefinition().load_file(
            SRC_ASSETS / "excel_definition.json"
        )
        output_to_excel_file(output_file, [], excel_definition)
        return output_file
    except Exception as e:
        print(e)
        return None


def _generate_jira_field_mapping_template(output_file: "Path") -> Optional[Path]:
    try:
        if generate_jira_field_mapping_file(output_file):
            return output_file
    except Exception as e:
        print(e)
    return None


def get_args_for_update_jira_info() -> Namespace:
    parser = ArgumentParser(
        description="Jira tool: Used to add/update jira url or access token.",
        formatter_class=ArgumentDefaultsHelpFormatter,
        allow_abbrev=False,
    )

    parser.add_argument(
        "--access_token",
        metavar="",
        type=str,
        required=False,
        help="Please follow the documentation to get your own access token.",
    )

    parser.add_argument(
        "--url",
        metavar="",
        type=str,
        required=False,
        help="Please provide the JIRA website url.",
    )

    return parser.parse_args()


def update_jira_info():
    try:
        args = get_args_for_update_jira_info()

        env_file: Path = SRC_ASSETS / ".env"

        if not env_file.exists():
            env_file.touch()

        # URL Part
        if args.url is not None:
            parsed_url: ParseResult = urlparse(str(args.url))

            if parsed_url.scheme not in ("https", "http"):
                print("Please check the jira url.")
            else:
                result, _, _ = set_key(
                    env_file,
                    key_to_set="JIRA_URL",
                    value_to_set=f"{parsed_url.scheme}://{parsed_url.netloc}",
                    quote_mode="never",
                )

                if result is True:
                    print("Add/Update jira url success!")
                else:
                    print("Add/Update jira url failed!")

        # ACCESS TOKEN Part
        if args.access_token is not None:
            access_token: str = str(args.access_token)

            if len(access_token.strip()) == 0 or access_token.isspace():
                print("Please check the access token.")
                sys.exit(1)
            else:
                result, _, _ = set_key(
                    env_file,
                    key_to_set="JIRA_ACCESS_TOKEN",
                    value_to_set=access_token,
                    quote_mode="never",
                )

                if result is True:
                    print("Add/Update jira access token success!")
                    sys.exit(0)
                else:
                    print("Add/Update jira access token failed!")
                    sys.exit(1)
    except Exception as e:
        print(e)
        sys.exit(1)



# -*- coding: utf-8 -*-
"""
This module is used to store excel column definition information.
"""
from __future__ import annotations

import re
from copy import deepcopy
from datetime import datetime
from json import loads
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import Any, List, Optional, Set, TypedDict, Union

from .milestone import Milestone
from .priority import Priority
from .utils import is_absolute_path_valid

__all__ = ["ExcelDefinition"]

RaiseRankingLevelScopeIndexValidationRule = re.compile(
    r"(^(\d{1,},){0,}\d{1,}$)|^\d{1,}-\d{1,}$"
)

# Load -> Parse -> Validate?


class PreProcessStep(TypedDict):
    name: str
    enabled: bool
    priority: Optional[int]
    config: dict


def parse_json_item_to_pre_process_step(json_item: Any) -> PreProcessStep:
    pre_process_step_name = ""
    pre_process_step_enabled = False
    pre_process_step_priority = 0
    pre_process_step_config = {}

    for key, value in json_item.items():
        if key.lower() == "name".lower():
            if value is None:
                raise ValueError("The pre-process step must have a name.")
            if isinstance(value, str):
                pre_process_step_name = value
            else:
                raise TypeError(
                    "The Name property type in the pre-process step should be string."
                )
        if key.lower() == "enabled".lower():
            if value is None:
                pre_process_step_enabled = False
            elif isinstance(value, bool):
                pre_process_step_enabled = value
            else:
                raise TypeError(
                    "The Enabled property type in the pre-process step should be boolean."
                )
        if key.lower() == "priority".lower():
            if value is None:
                pre_process_step_priority = 0
            elif isinstance(value, int):
                pre_process_step_priority = value
            else:
                raise TypeError(
                    "The Priority property type in the pre-process step should be integer."
                )
        if key.lower() == "config".lower():
            pre_process_step_config = {}
            if value is not None and isinstance(value, dict):
                for name in value:
                    # More config support.
                    if name.lower() == "JiraStatuses".lower():
                        pre_process_step_config["JiraStatuses"] = value.get(name, None)

    return PreProcessStep(
        name=pre_process_step_name,
        enabled=pre_process_step_enabled,
        priority=pre_process_step_priority,
        config=pre_process_step_config,
    )


class SortStrategy(TypedDict):
    name: str
    priority: Optional[int]
    enabled: bool
    config: dict


def parse_json_item_to_sort_strategy(json_item: Any) -> SortStrategy:
    strategy_name = ""
    strategy_priority = 0
    strategy_enabled = False
    strategy_config = {}

    for key, value in json_item.items():
        if key.lower() == "name".lower():
            if value is None:
                raise ValueError("The sort strategy must have a name.")
            if isinstance(value, str):
                strategy_name = value
            else:
                raise TypeError(
                    "The Name property type in the sort strategy should be string."
                )
        if key.lower() == "priority".lower():
            if value is None:
                strategy_priority = 0
            elif isinstance(value, int):
                strategy_priority = value
            else:
                raise TypeError(
                    "The Priority property type in the sort strategy should be integer."
                )
        if key.lower() == "enabled".lower():
            if value is None:
                strategy_enabled = False
            elif isinstance(value, bool):
                strategy_enabled = value
            else:
                raise TypeError(
                    "The Enabled property type in the sort strategy should be boolean."
                )
        if key.lower() == "config".lower():
            strategy_config = {}
            if value is not None and isinstance(value, dict):
                for name in value:
                    # More config support.
                    if name.lower() == "ParentScopeIndexRange".lower():
                        strategy_config[
                            name
                        ] = ExcelDefinition.parse_raise_ranking_level_scope_index_expression(
                            value.get(name, None)
                        )

    return SortStrategy(
        name=strategy_name,
        priority=strategy_priority,
        enabled=strategy_enabled,
        config=strategy_config,
    )


class ExcelDefinitionColumn(TypedDict):
    index: int
    name: str
    type: Optional[type]
    require_sort: bool
    sort_order: bool
    scope_require_sort: bool
    scope_sort_order: bool
    inline_weights: int
    raise_ranking: int
    scope_raise_ranking: int
    jira_field_mapping: Optional[dict[str, str]]


def parse_json_item_to_excel_definition_column(json_item: Any) -> ExcelDefinitionColumn:
    column_index = 0
    column_name = ""
    column_type = None
    column_require_sort = False
    column_sort_order = False
    column_scope_require_sort = False
    column_scope_sort_order = False
    column_inline_weights = -1
    column_raise_ranking = -1
    column_scope_raise_ranking = -1
    column_jira_field_mapping = None

    for key, value in json_item.items():
        if key.lower() == "index".lower():
            if value is None:
                raise ValueError("Column definition must has an index.")
            if isinstance(value, int):
                column_index = value
            else:
                raise TypeError(
                    "The Index property type in the column definition is not integer."
                )
        elif key.lower() == "name".lower():
            if value is None:
                raise ValueError("Column definition must has a name.")
            if isinstance(value, str):
                column_name = value
            else:
                raise TypeError(
                    "The Name property type in the column definition should be string."
                )
        elif key.lower() == "type".lower():
            column_type = ExcelDefinition.convert_str_to_type(value)
        elif key.lower() == "RequireSort".lower():
            column_require_sort = value
        elif key.lower() == "SortOrder".lower():
            column_sort_order = value
        elif key.lower() == "ScopeRequireSort".lower():
            column_scope_require_sort = value
        elif key.lower() == "ScopeSortOrder".lower():
            column_scope_sort_order = value
        elif key.lower() == "InlineWeights".lower():
            column_inline_weights = value
        elif key.lower() == "RaiseRanking".lower():
            column_raise_ranking = value
        elif key.lower() == "ScopeRaiseRanking".lower():
            column_scope_raise_ranking = value
        elif key.lower() == "JiraFieldMapping".lower():
            column_jira_field_mapping = value

    return ExcelDefinitionColumn(
        index=column_index,
        name=column_name,
        type=column_type,
        require_sort=column_require_sort,
        sort_order=column_sort_order,
        scope_require_sort=column_scope_require_sort,
        scope_sort_order=column_scope_sort_order,
        inline_weights=column_inline_weights,
        raise_ranking=column_raise_ranking,
        scope_raise_ranking=column_scope_raise_ranking,
        jira_field_mapping=column_jira_field_mapping,
    )


class ExcelDefinition:
    def __init__(self) -> None:
        self.columns: list[ExcelDefinitionColumn] = []
        self.sort_strategies: list[SortStrategy] = []
        self.pre_process_steps: list[PreProcessStep] = []

    def load(self, content: str) -> "ExcelDefinition":
        """
        Load json string to generate the excel definition

        :param content:
            JSON string content
        """

        if content is None:
            raise ValueError("There is no content in the excel definition file.")

        try:
            raw_data = loads(content)
        except JSONDecodeError as e:
            raise SyntaxError(
                f"The structure of excel definition file is wrong. Hint: {e.msg} in line {e.lineno}:{e.colno}."
            ) from e

        parse_errors = []

        if len(raw_data) > 0:
            if raw_data[0].get("PreProcessSteps", None) is not None:
                for item in raw_data[0]["PreProcessSteps"]:
                    try:
                        self.pre_process_steps.append(
                            parse_json_item_to_pre_process_step(item)
                        )
                    except (TypeError, ValueError) as e:
                        parse_errors.append(e.args[0])

            if raw_data[0].get("SortStrategies", None) is not None:
                for item in raw_data[0]["SortStrategies"]:
                    try:
                        self.sort_strategies.append(
                            parse_json_item_to_sort_strategy(item)
                        )
                    except (TypeError, ValueError) as e:
                        parse_errors.append(e.args[0])

        if len(raw_data) >= 1:
            for item in raw_data[1]["Columns"]:
                try:
                    self.columns.append(
                        parse_json_item_to_excel_definition_column(item)
                    )
                except (TypeError, ValueError) as e:
                    parse_errors.append(e.args[0])

        if parse_errors:
            # Avoid duplicate error messages.
            parse_error_message = "\n".join(
                [f"{index + 1}. {err}" for index, err in enumerate(set(parse_errors))]
            )
            raise SyntaxError(
                f"The excel definition file has below issues need to be fixed:\n{parse_error_message}"
            )

        return self

    @staticmethod
    def parse_raise_ranking_level_scope_index_expression(
        expression: Union[Any, None],
    ) -> Optional[Set[int]]:
        if expression is None or not isinstance(expression, str):
            return None
        if len(expression) == 0 or expression.isspace():
            return set()
        if (
            RaiseRankingLevelScopeIndexValidationRule.fullmatch(
                "".join(expression.split(" "))
            )
            is None
        ):
            return None  # None means invalid, since we don't have the parse procedure.
        if "-" in expression:
            begin = int(expression.split("-")[0])
            end = int(expression.split("-")[1])
            if begin < end:
                return set(i for i in range(begin, end + 1))
            return set(i for i in range(end, begin + 1))
        return set(int(i) for i in expression.split(","))

    def load_file(self, file: Union[str, Path]) -> "ExcelDefinition":
        """
        Load json file to generate the excel definition

        :param file:
            JSON file location
        """

        if not is_absolute_path_valid(file):
            raise FileNotFoundError(
                f"Please make sure the excel definition file exist and the path should be absolute. File: {file}."
            )

        with open(file=file, mode="r", encoding="utf-8") as table_definition_file:
            try:
                self.load(table_definition_file.read())
            finally:
                table_definition_file.close()

        return self

    def validate(self) -> "List":
        return (
            self._validate_pre_process_steps()
            + self._validate_sort_strategies()
            + self._validate_column_definitions()
        )

    def _validate_pre_process_steps(self) -> "List[str]":
        invalid_definitions = []

        # Validate PreProcessSteps
        pre_process_step_priorities: List[int] = []
        for pre_process_step in self.pre_process_steps:
            if pre_process_step["name"].isspace() or len(pre_process_step["name"]) == 0:
                invalid_definitions.append("The PreProcessStep name is invalid.")
                # If strategy name is invalid, no need to check more.
                continue

            if (
                pre_process_step["priority"] is None
                or not isinstance(pre_process_step["priority"], int)
                or pre_process_step["priority"] < 0
            ):
                invalid_definitions.append(
                    f"The pre-process step priority is invalid. PreProcessStep: {pre_process_step['name']}"
                )
            else:
                if pre_process_step["priority"] in pre_process_step_priorities:
                    invalid_definitions.append(
                        f"The pre-process step priority is duplicate. PreProcessStep: {pre_process_step['name']}"
                    )
                else:
                    pre_process_step_priorities.append(pre_process_step["priority"])

            if "JiraStatuses".lower() in [
                config_name.lower() for config_name in pre_process_step["config"].keys()
            ] and (
                pre_process_step["name"].lower()
                != "FilterOutStoryBasedOnJiraStatus".lower()
            ):
                invalid_definitions.append(
                    f"Only FilterOutStoryBasedOnJiraStatus step support JiraStatuses config. PreProcessStep: {pre_process_step['name']}."
                )

            if "JiraStatuses".lower() in [
                config_name.lower() for config_name in pre_process_step["config"].keys()
            ] and not isinstance(pre_process_step["config"]["JiraStatuses"], list):
                invalid_definitions.append(
                    f"The format of the Jira Statuses is invalid. PreProcessStep: {pre_process_step['name']}. Supported format like: ['CLOSED', 'PENDING RELEASE']."
                )

        return invalid_definitions

    def _validate_sort_strategies(self) -> "List[str]":
        invalid_definitions = []

        # Validate Strategies
        strategy_priorities: List[int] = []
        for strategy in self.sort_strategies:
            if strategy["name"].isspace() or len(strategy["name"]) == 0:
                invalid_definitions.append("The strategy name is invalid.")
                # If strategy name is invalid, no need to check more.
                continue

            if (
                strategy["priority"] is None
                or not isinstance(strategy["priority"], int)
                or strategy["priority"] < 0
            ):
                invalid_definitions.append(
                    f"The strategy priority is invalid. Strategy: {strategy['name']}"
                )
            else:
                if strategy["priority"] in strategy_priorities:
                    invalid_definitions.append(
                        f"The strategy priority is duplicate. Strategy: {strategy['name']}"
                    )
                else:
                    strategy_priorities.append(strategy["priority"])

            if "ParentScopeIndexRange".lower() in [
                config_name.lower() for config_name in strategy["config"].keys()
            ] and (
                strategy["name"].lower() != "SortOrder".lower()
                and strategy["name"].lower() != "RaiseRanking".lower()
            ):
                invalid_definitions.append(
                    f"Only RaiseRanking and SortOrder strategy support ParentScopeIndexRange config. Strategy: {strategy['name']}."
                )

            if "ParentScopeIndexRange".lower() in [
                config_name.lower() for config_name in strategy["config"].keys()
            ] and not isinstance(strategy["config"]["ParentScopeIndexRange"], set):
                invalid_definitions.append(
                    f"The format of the Parent Level Index Range is invalid. Strategy: {strategy['name']}. Supported format strings like: 1-20 or 20,30 or empty string."
                )

        return invalid_definitions

    def _validate_column_definitions(self) -> "List[str]":
        invalid_definitions = []

        # Validate the Columns
        exist_story_id_column = False
        exist_indexes = []
        exist_inline_weights = []
        for column in self.get_columns():
            column_index: int = column["index"]
            column_name: str = column["name"]
            column_type: Optional[type] = column["type"]
            column_require_sort: bool = column["require_sort"]
            column_sort_order: bool = column["sort_order"]
            column_scope_require_sort: bool = column["scope_require_sort"]
            column_scope_sort_order: bool = column["scope_sort_order"]
            column_inline_weights: int = column["inline_weights"]
            column_raise_ranking: int = column["raise_ranking"]
            column_scope_raise_ranking: int = column["scope_raise_ranking"]
            column_jira_field_mapping: Optional[dict] = column["jira_field_mapping"]

            # Check Name cannot be empty
            if len(column_name) == 0:
                invalid_definitions.append(
                    f"Column name cannot be empty. Index: {column_index}"
                )
                continue

            if column_name.lower() == "StoryId".lower():
                exist_story_id_column = True

            # Check Missing/Duplicate Index
            if not isinstance(column_index, int):
                invalid_definitions.append(
                    f"Column Index can only be number. Column: {column_name}"
                )
            elif column_index is None:
                invalid_definitions.append(f"Missing Index. Column: {column_name}")
            elif column_index in exist_indexes:
                invalid_definitions.append(f"Duplicate Index. Column: {column_name}")
            else:
                exist_indexes.append(column_index)
            # Check Property Type
            if column_type not in (
                str,
                bool,
                datetime,
                Priority,
                Milestone,
                float,
            ):
                invalid_definitions.append(
                    f"Invalid Column Type. Column: {column_name}"
                )

            # Check Sort
            if not isinstance(column_require_sort, bool):
                invalid_definitions.append(
                    f"Require Sort can only be True/False. Column: {column_name}"
                )

            if not isinstance(column_sort_order, bool):
                invalid_definitions.append(
                    f"Sort Order can only be True(Descending)/False(Ascending). Column: {column_name}"
                )

            # Check Sort
            if not isinstance(column_scope_require_sort, bool):
                invalid_definitions.append(
                    f"Scope Require Sort can only be True/False. Column: {column_name}"
                )

            if not isinstance(column_scope_sort_order, bool):
                invalid_definitions.append(
                    f"Scope Sort Order can only be True(Descending)/False(Ascending). Column: {column_name}"
                )

            # Check InlineWeights
            if not isinstance(column_inline_weights, int):
                invalid_definitions.append(
                    f"Inline Weights can only be number. Column: {column_name}"
                )
            else:
                if (
                    column_inline_weights > 0
                    and column_inline_weights in exist_inline_weights
                ):
                    invalid_definitions.append(
                        f"Duplicate Inline Weights. Currently only support different line weights. Column: {column_name}"
                    )
                exist_inline_weights.append(column_inline_weights)

            # Check RaiseRanking
            if not isinstance(column_raise_ranking, int):
                invalid_definitions.append(
                    f"Raise Ranking can only be number. Column: {column_name}"
                )
            else:
                # Check Support RaiseRanking or not
                if column_type not in (bool,) and column_raise_ranking > 0:
                    invalid_definitions.append(
                        f"Column do not support Raise Ranking feature. Column: {column_name}"
                    )

            if not isinstance(column_scope_raise_ranking, int):
                invalid_definitions.append(
                    f"Scope Raise Ranking can only be number. Column: {column_name}"
                )
            else:
                # Check Support RaiseRanking or not
                if column_type not in (bool,) and column_scope_raise_ranking > 0:
                    invalid_definitions.append(
                        f"Column do not support Scope Raise Ranking feature. Column: {column_name}"
                    )

            if column_jira_field_mapping is None:
                continue
            if column_jira_field_mapping is not None and not isinstance(
                column_jira_field_mapping, dict
            ):
                invalid_definitions.append(
                    f"Jira Field Mapping can only be dictionary. Column: {column_name}"
                )
            else:
                jira_field_name = column_jira_field_mapping.get("name", None)
                if jira_field_name is None or jira_field_name.isspace():
                    invalid_definitions.append(
                        f"Jira Field Mapping has the invalid name. Column: {column_name}"
                    )
                jira_field_path = column_jira_field_mapping.get("path", None)
                if jira_field_path is None or jira_field_path.isspace():
                    invalid_definitions.append(
                        f"Jira Field Mapping has the invalid path. Column: {column_name}"
                    )

        if len(self.columns) > 0 and exist_story_id_column is False:
            invalid_definitions.append(
                "Must have a column named StoryId so that program can identify the record."
            )

        if len(invalid_definitions) == 0:
            self.columns.sort(key=lambda c: c["index"], reverse=False)

            if len(self.columns) > 0 and (
                self.columns[0]["index"] != 1
                or self.columns[len(self.columns) - 1]["index"] != len(self.columns)
            ):
                invalid_definitions.append(
                    "Column Index must be in continuation and starts from 1."
                )

        return invalid_definitions

    @staticmethod
    def convert_str_to_type(type_str: str) -> Optional[type]:
        if type_str is None or not isinstance(type_str, str):
            return None
        type_str = str(type_str).strip().lower()
        if type_str.lower() == "str":
            return str
        if type_str.lower() == "bool":
            return bool
        if type_str.lower() == "datetime":
            return datetime
        if type_str.lower() == "priority":
            return Priority
        if type_str.lower() == "milestone":
            return Milestone
        # Currently, only support float/double
        if type_str.lower() == "number":
            return float
        return None

    def __iter__(self):
        for item in self.columns:
            yield item

    def get_columns(self) -> "List[ExcelDefinitionColumn]":
        return deepcopy(self.columns)

    def get_columns_name(self) -> "List[Optional[str]]":
        return [item["name"] for item in self.columns]

    @property
    def max_column_index(self) -> int:
        return self.columns[len(self.columns) - 1]["index"]

    def get_sort_strategies(self, enabled: bool = True) -> "List[SortStrategy]":
        result: list[SortStrategy] = []
        for sort_strategy in self.sort_strategies:
            if sort_strategy["enabled"] == enabled:
                result.append(deepcopy(sort_strategy))
        result.sort(key=_sort_priority_map, reverse=False)
        return result

    def get_pre_process_steps(self, enabled: bool = True) -> "List[PreProcessStep]":
        result: list[PreProcessStep] = []
        for pre_process_step in self.pre_process_steps:
            if pre_process_step["enabled"] == enabled:
                result.append(deepcopy(pre_process_step))
        result.sort(key=_sort_priority_map, reverse=False)
        return result

    def total_count(self):
        return len(self.columns)


def _sort_priority_map(item: Any) -> int:
    if item["priority"] is None or not isinstance(item["priority"], int):
        return 0
    return item["priority"]



import pathlib
from json import loads
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import List, Union

__all__ = ["SprintScheduleStore"]


class SprintScheduleStore:
    def __init__(self) -> None:
        self.store: List[tuple] = []

    def load(self, content: str):
        """
        Load json string to generate the priority list

        :param content:
            JSON string content
        """
        if content is None:
            return

        try:
            raw_data = loads(content)
        except JSONDecodeError as e:
            raise SyntaxError(
                f"The structure of sprint schedule file is wrong. Hint: {e.msg} in line {e.lineno}:{e.colno}."
            ) from e

        priority = 0
        sprints = []
        for item in raw_data:
            for key, value in item.items():
                if key.lower() in "priority":
                    if value is None or not isinstance(value, int):
                        # Just skip invalid items.
                        continue
                    priority = value
                if key.lower() in "sprints":
                    if value is None or not isinstance(value, list):
                        continue
                    for sprint in value:
                        if len(sprint) > 0:
                            sprints.append(sprint)

            for sprint in sprints:
                self.store.append((sprint, priority))
            sprints.clear()
            priority = 0

    def load_file(self, file: Union[str, Path]):
        """
        Load json file to generate the excel definition

        :param file:
            JSON file location
        """

        if (
            file is None
            or not pathlib.Path(file).is_absolute()
            or not pathlib.Path(file).exists()
        ):
            raise FileNotFoundError(
                f"Please make sure the sprint schedule file exist and the path should be absolute. File: {file}."
            )

        with open(file=file, mode="r", encoding="utf-8") as schedule_file:
            try:
                self.load(schedule_file.read())
            finally:
                schedule_file.close()

    def get_priority(self, sprint: str) -> int:
        for item in self.store:
            if sprint.upper() in item[0].upper():
                return item[1]
        return 0

    def total_count(self):
        return len(self.store)



import pathlib

from pytest import raises

from jira_assistant.excel_definition import ExcelDefinition

HERE = pathlib.Path(__file__).resolve().parent
SRC_ASSETS = HERE.parent / "src/jira_assistant/assets"
TEST_ASSETS = HERE / "files"


class TestExcelDefinition:
    def test_load_happy_path(self):
        excel_definition_filename = SRC_ASSETS / "excel_definition.json"

        store = ExcelDefinition()
        with open(excel_definition_filename, encoding="utf-8") as file:
            store.load(file.read())
        assert store.total_count() > 0

    def test_load_using_none_input(self):
        store = ExcelDefinition()
        content: str = None  # type: ignore
        with raises(ValueError) as err:
            store.load(content)
        assert "There is no content in the excel definition file." in str(err.value)

    def test_load_file(self):
        excel_definition_filename = SRC_ASSETS / "excel_definition.json"
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)
        assert store.total_count() > 0

    def test_load_file_none(self):
        store = ExcelDefinition()
        file: str = None  # type: ignore
        with raises(FileNotFoundError) as err:
            store.load_file(file)
        assert (
            "Please make sure the excel definition file exist and the path should be absolute"
            in str(err.value)
        )

    def test_iter(self):
        excel_definition_filename = SRC_ASSETS / "excel_definition.json"
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)
        items = []
        for item in store:
            items.append(item)
        assert len(items) > 0

    def test_validate(self):
        excel_definition_filename = SRC_ASSETS / "excel_definition.json"
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 0

    def test_validate_missing_story_id(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_missing_story_id.json"
        )
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 1

    def test_validate_duplicate_index(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_duplicate_index.json"
        )
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 1

    def test_validate_duplicate_inline_weights(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_duplicate_inline_weights.json"
        )
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 1

    def test_validate_invalid_name(self):
        excel_definition_filename = TEST_ASSETS / "excel_definition_invalid_name.json"
        store = ExcelDefinition()
        with raises(SyntaxError) as err:
            store.load_file(excel_definition_filename)

        assert (
            "The excel definition file has below issues need to be fixed"
            in err.value.args[0]
        )
        assert (
            "The Name property type in the column definition should be string."
            in err.value.args[0]
        )

    def test_validate_invalid_raise_ranking(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_invalid_raise_ranking.json"
        )
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 1

    def test_validate_invalid_require_sort(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_invalid_require_sort.json"
        )
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 2

    def test_validate_invalid_type(self):
        excel_definition_filename = TEST_ASSETS / "excel_definition_invalid_type.json"
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 2

    def test_validate_invalid_index(self):
        excel_definition_filename = TEST_ASSETS / "excel_definition_invalid_index.json"
        store = ExcelDefinition()
        with raises(SyntaxError) as err:
            store.load_file(excel_definition_filename)

        assert (
            "The excel definition file has below issues need to be fixed"
            in err.value.args[0]
        )
        assert (
            "The Index property type in the column definition is not integer."
            in err.value.args[0]
        )

    def test_validate_index_not_continuation(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_index_not_continuation.json"
        )
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 1

    def test_validate_strategy_duplicate_priority(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_strategy_duplicate_priority.json"
        )
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 1

    def test_validate_strategy_invalid_name(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_strategy_invalid_name.json"
        )
        store = ExcelDefinition()
        store.load_file(excel_definition_filename)

        validation_result = store.validate()

        assert len(validation_result) == 1

    def test_validate_invalid_structure(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_invalid_structure.txt"
        )
        store = ExcelDefinition()

        with raises(Exception) as err:
            store.load_file(excel_definition_filename)

        assert (
            "The structure of excel definition file is wrong. Hint: Expecting ',' delimiter in line 49:1"
            in err.value.args[0]
        )

    def test_validate_invalid_pre_process_step_name(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_invalid_pre_process_step_name.json"
        )
        store = ExcelDefinition()

        with raises(SyntaxError) as err:
            store.load_file(excel_definition_filename)

        assert (
            "The excel definition file has below issues need to be fixed"
            in err.value.args[0]
        )
        assert "The pre-process step must have a name." in err.value.args[0]

    def test_validate_invalid_pre_process_step_priority(self):
        excel_definition_filename = (
            TEST_ASSETS / "excel_definition_invalid_pre_process_step_priority.json"
        )
        store = ExcelDefinition()

        store.load_file(excel_definition_filename)
        validation_result = store.validate()

        assert len(validation_result) == 1



import pathlib
from os import environ, remove

from mock_server import mock_jira_requests
from requests_mock import Mocker
from utils import read_stories_from_excel

from jira_assistant.assistant import (
    generate_jira_field_mapping_file,
    run_steps_and_sort_excel_file,
)
from jira_assistant.excel_definition import ExcelDefinition
from jira_assistant.excel_operation import read_excel_file
from jira_assistant.jira_client import JiraClient
from jira_assistant.sprint_schedule import SprintScheduleStore
from jira_assistant.story import compare_story_based_on_inline_weights

HERE = pathlib.Path(__file__).resolve().parent
SRC_ASSETS: pathlib.Path = HERE.parent / "src/jira_assistant/assets"


class TestAssistant:
    def test_run_steps_and_sort_excel_file(self):
        with Mocker(
            real_http=False, case_sensitive=False, adapter=mock_jira_requests()
        ):
            run_steps_and_sort_excel_file(
                HERE / "files/happy_path.xlsx",
                HERE / "files/happy_path_sorted.xlsx",
                excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
                sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            )

            excel_definition = ExcelDefinition()
            excel_definition.load_file(SRC_ASSETS / "excel_definition.json")
            sprint_schedule = SprintScheduleStore()
            sprint_schedule.load_file(SRC_ASSETS / "sprint_schedule.json")

            _, stories = read_excel_file(
                HERE / "files/happy_path_sorted.xlsx", excel_definition, sprint_schedule
            )

            assert len(stories) == 8

            jira_client = JiraClient(environ["JIRA_URL"], environ["JIRA_ACCESS_TOKEN"])

            noneed_sort_statuses = [
                "SPRINT COMPLETE",
                "PENDING RELEASE",
                "PRODUCTION TESTING",
                "CLOSED",
            ]

            jira_fields = [
                {
                    "name": "domain",
                    "jira_name": "customfield_15601",
                    "jira_path": "customfield_15601.value",
                },
                {"name": "status", "jira_name": "status", "jira_path": "status.name"},
            ]

            for i in range(len(stories) - 1):
                story_id_0 = stories[i]["storyId"].lower().strip()
                story_id_1 = stories[i + 1]["storyId"].lower().strip()
                query_result = jira_client.get_stories_detail(
                    [story_id_0, story_id_1], jira_fields
                )
                if (
                    query_result[story_id_0]["status"].upper()
                    not in noneed_sort_statuses
                    and query_result[story_id_1]["status"].upper()
                    not in noneed_sort_statuses
                ):
                    assert (
                        compare_story_based_on_inline_weights(
                            stories[i], stories[i + 1]
                        )
                        >= 0
                    )

            remove(HERE / "files/happy_path_sorted.xlsx")

    def test_run_steps_and_sort_excel_file_with_empty_excel_file(self):
        with Mocker(
            real_http=False, case_sensitive=False, adapter=mock_jira_requests()
        ):
            run_steps_and_sort_excel_file(
                HERE / "files/empty_excel.xlsx",
                HERE / "files/empty_excel_sorted.xlsx",
                excel_definition_file=str(SRC_ASSETS / "excel_definition.json"),
                sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            )

    def test_run_steps_and_sort_excel_file_with_raise_ranking_file(self):
        with Mocker(
            real_http=False, case_sensitive=False, adapter=mock_jira_requests()
        ):
            run_steps_and_sort_excel_file(
                HERE / "files/happy_path.xlsx",
                HERE / "files/happy_path_sorted.xlsx",
                excel_definition_file=str(
                    HERE / "files/excel_definition_with_raise_ranking.json"
                ),
                sprint_schedule_file=str(SRC_ASSETS / "sprint_schedule.json"),
            )

            stories = read_stories_from_excel(
                HERE / "files/happy_path_sorted.xlsx",
                SRC_ASSETS / "excel_definition.json",
                SRC_ASSETS / "sprint_schedule.json",
            )

            false_value_begin = False
            for story in stories:
                if story["isThisAHardDate"] is True:
                    continue
                if story["isThisAHardDate"] is False and false_value_begin is False:
                    false_value_begin = True
                    continue
                if story["isThisAHardDate"] is True and false_value_begin is True:
                    raise AssertionError

            remove(HERE / "files/happy_path_sorted.xlsx")

    def test_generate_jira_field_mapping_file(self):
        with Mocker(
            real_http=False, case_sensitive=False, adapter=mock_jira_requests()
        ):
            output_file: pathlib.Path = HERE / "temp/jira_field_mapping.json"

            assert generate_jira_field_mapping_file(output_file) is True
            assert output_file.exists() is True

            remove(output_file)
